#include "grid2d.h"
#include "Distribution/diracDistribution.h"
#include "Distribution/discreteDistribution.h"
#include "transitionStructure/sparseMatrix.h"
#include "transitionStructure/multiDimHomTransition.h"
#include <stdlib.h>

/**
 * @brief Constructor for a grid2d chain with infinite state space
 * @author	Alain Jean-Marie
 * @param p	right-jump probability
 * @param q	left-jump probability
 * @return	an object of type grid2d, with initial state set arbitrarily to (0,0)
 */
grid2d::grid2d( double* p, double* q )
  : markovChain( INFINITE_STATE_SPACE_SIZE, DISCRETE )
{
  _nbDims = 2;
  _p = p;
  _q = q;
  _r = 1.0 - ( _p[0] + _p[1] + _q[0] + q[1] ); // should test the values of these three probas
  _stateSpace = (double*)NULL;

  int sz[] = { INFINITE_STATE_SPACE_SIZE, INFINITE_STATE_SPACE_SIZE };
  _dimSize = sz;
  _generator = new multiDimHomTransition( 2, sz, _p, _q );

  _initDistribution = new diracDistribution( 0 );

  // creating and filling the _mu array for consistency. Probably not very useful.
  _mu = (int*)malloc( _nbDims*sizeof(int) );
  for ( int d = 0; d < _nbDims; d++ ) {
    _mu[d] = INFINITE_STATE_SPACE_SIZE;
  }

}

/**
 * @brief Constructor for a grid2d chain
 * @author	Alain Jean-Marie
 * @param sz sizes of the state space in each dimension
 * @param p	right-jump probabilities
 * @param q	left-jump probabilities
 * @return	an object of type markovChain, with initial state set
 		arbitrarily to 0.
 */
grid2d::grid2d( int* sz, double* p, double* q )
  : markovChain( sz[0]*sz[1], DISCRETE )
{
  _nbDims = 2;
  _dimSize = sz;
  _p = p;
  _q = q;
  _r = 1.0 - ( _p[0] + _p[1] + _q[0] + q[1] ); // should test the values of these three probas

  // _stateSpaceSize = sz[0]*sz[1] is already set by super constructor
  _stateSpace = (double*)malloc( _stateSpaceSize * sizeof(double) );
  for ( int i = 0; i < _stateSpaceSize; i++ ) {
    _stateSpace[i] = i;
  }

  _generator = new multiDimHomTransition( 2, sz, _p, _q );

  _initDistribution = new diracDistribution( 0 );

  _mu = (int*)malloc( _nbDims*sizeof(int) );
  _mu[_nbDims-1] = 1;
  for ( int d = _nbDims-2; d >= 0; d-- ) {
    _mu[d] = _mu[d+1] * _dimSize[d+1];
  }

}

/**
 * @brief Destructor for a grid2d chain
 * @author	Alain Jean-Marie
 */
grid2d::~grid2d()
{
  free( _stateSpace );
  free( _mu );
  // delete _initDistribution; // NO: is done in mother class desctructor
}

/**
 * @brief Instantiation of the generator for the markovChain ancestor
 * @author	Alain Jean-Marie
 */
void grid2d::makeMarkovChain()
{
  sparseMatrix* newgen = new sparseMatrix( _stateSpaceSize );
  newgen->setType( DISCRETE );

  for ( int i = 0; i < _dimSize[0]; i++ ) {
    for ( int j = 0; j < _dimSize[1]; j++ ) {
      int stateNum = j*_dimSize[0]+i;
      newgen->addToEntry( stateNum, (i==_dimSize[0]-1?stateNum:stateNum+1), _p[0] );
      newgen->addToEntry( stateNum, (i==0?stateNum:stateNum-1), _q[0] );
      newgen->addToEntry( stateNum, (j==_dimSize[1]?stateNum:stateNum+_dimSize[0]), _p[1] );
      newgen->addToEntry( stateNum, (j==0?stateNum:stateNum-_dimSize[0]), _q[1] );
      newgen->addToEntry( stateNum, stateNum, _r );
    }
  }

  delete _generator;
  _generator = newgen;

}

/**
 * @brief Performs Monte-Carlo sampling of hitting times
 * @author Alain Jean-Marie
 * @param iState the initial state
 * @param hittingSet a boolean array marking with true the states in the hitting set
 * @param nbSamples number of samples to obtain
 * @return an array containing the samples
 */
int* grid2d::simulateHittingTime( int iState, bool *hittingSet, int nbSamples, int tMax )
{
  multiDimHomTransition* nGen = dynamic_cast<multiDimHomTransition*>(_generator);
  discreteDistribution* jump = nGen->getJumpDistribution();
  int* res = (int*)malloc( nbSamples*sizeof(int) );
  int stateBuffer[2];

  for ( int i = 0; i < nbSamples; i++ ) {
    int curState = iState;
    stateBuffer[0] = curState / _dimSize[1];
    stateBuffer[1] = curState % _dimSize[1];
    int hitT = 0;

    // fprintf( stdout, "Traj #%d: %d", i, curState );
    // bool inRun = false;
    // int runL = 0;

    while( !hittingSet[curState] && ( hitT < tMax ) ) {
      hitT++;
      int theJump = jump->sample();
      // fprintf( stderr, "%d\n", theJump );
      if ( theJump != 0 ) {
        int theDim = abs(theJump)-1;
        int theDir = theJump/abs(theJump);
        if ( ( stateBuffer[theDim] + theDir >= 0 ) && ( stateBuffer[theDim] + theDir < _dimSize[theDim] ) ) {
          // accept the jump
          curState += theDir*_mu[theDim];
          stateBuffer[theDim] += theDir;
        }
      }
//      if ( inRun ) {
//        if ( curState == 12 ) {
//          runL++;
//        }
//        else {
//          // end of run
//          inRun = false;
//          fprintf( stderr, "Run12: %d\n", runL );
//        }
//      }
//      else if ( curState == 12 ) {
//        inRun = true;
//        runL = 1;
//      }
      // fprintf( stdout, " %d", curState );
    }
    res[i] = hitT;
    // fprintf( stdout, "\n");
  }

  // cleanup
  delete jump;

  //finished
  return res;
}

/**
 * @brief Writes a grid2d chain in files with some specified format.
 * Supported formats:
 * - Xborne: both Rii and Cuu files are written
 * - Psi, Ers
 * @author Alain Jean-Marie
 * @param format format/language to be used
 * @param modelName the name of the model, usually used as prefix for the different files
 */

void grid2d::write( string format, string modelName )
{
  // To be done: check that chain is finite
  if ( format == "XBORNE" ) {
    // write "size", "Rii", "Cuu" and "cd" files.
    string name = modelName + ".sz";
    FILE* out = fopen( name.c_str(), "w" );
    if ( NULL != out ) {
      fprintf( out, "%16d\n", 5*_stateSpaceSize - 2*(_stateSpaceSize/_dimSize[0]+_stateSpaceSize/_dimSize[1]) );
      fprintf( out, "%16d\n", _stateSpaceSize );
      fprintf( out, "%16d\n", 2 );
      fclose(out);
    }
    else {
      cerr << "Error in grid2d::write(): Cannot open file \"" 
	   << name << "\" for writing." << endl;
    }
    
    name = modelName + ".Rii";
    out = fopen( name.c_str(), "w" );
    if ( NULL != out ) {
      multiDimHomTransition* nGen = dynamic_cast<multiDimHomTransition*>(_generator);
      nGen->write( out, "XBORNE" );
      fclose(out);
    }
    else {
      cerr << "Error in grid2d::write(): Cannot open file \"" 
	   << name << "\" for writing." << endl;
    }

    name = modelName + ".Cuu";
    out = fopen( name.c_str(), "w" );
    if ( NULL != out ) {
      // _generator->writeTranspose( out, "XBORNE" );
      fclose(out);
    }
    else {
      cerr << "Error in grid2d::write(): Cannot open file \""
           << name << "\" for writing." << endl;
    }

    name = modelName + ".cd";
    out = fopen( name.c_str(), "w" );
    if ( NULL != out ) {
      // list states in lexicographic order
      for ( int i = 0; i < _stateSpaceSize; i++ ) {
        int x = i / _dimSize[1];
        int y = i % _dimSize[1];
        fprintf( out, "%10d%10d%10d", i, x, y );
        fprintf( out, "\n" );
      }
      fclose(out);
    }
    else {
      cerr << "Error in grid2d::write(): Cannot open file \""
           << name << "\" for writing." << endl;
    }
  }
  else if ( format == "MARCA" ) {

    string name = modelName + ".marca";
    FILE* out = fopen( name.c_str(), "w" );
    if ( NULL != out ) {
      _generator->write( out, "MARCA");
      fclose(out);
    }
    else {
      cerr << "Error in grid2d::write(): Cannot open file \""
	   << name << "\" for writing." << endl;
    }
  }
  else if ( format == "Ers" ) {

    string name = modelName + ".mcl";
    FILE* out = fopen( name.c_str(), "w" );
    if ( NULL != out ) {
      _generator->write( out, "Ers" );
      fclose(out);
    }
    else {
      cerr << "Error in grid2d::write(): Cannot open file \"" 
	   << name << "\" for writing." << endl;
    }
  }
  else if ( format == "Maple" ) {

    string name = modelName + ".map";
    FILE* out = fopen( name.c_str(), "w" );
    if ( NULL != out ) {
      fprintf( out, "P_%s := Matrix( %d, {\n", modelName.c_str(), _stateSpaceSize );
      _generator->write( out, "Maple" );
      fprintf( out, "}, fill=0 );\n");
      fclose(out);
    }
    else {
      cerr << "Error in grid2d::write(): Cannot open file \""
       << name << "\" for writing." << endl;
    }
  }
  else {
      cerr << "Error in grid2d::write(): Format \""
       << format << "\" not handled." << endl;
  }
}
