#include "layeredStateSpace.h"
#include <stdlib.h>
#include <iostream>

using namespace std;

// constructor for boxes with the SW corner at (0,...,0)
layeredStateSpace::layeredStateSpace(int N, int E1, int E2, int M, int nu )
{
  // specific class variables
  _N = N;
  _E1 = E1;
  _E2 = E2;
  _M = M;
  _nu = nu;

  _S = _M * ( 1 + _E2 );
  _L0 = _S * ( 1 + _E1 );
  _L = _S * ( 1 + _nu * _E1 );

  // generic class variables
  _isSimple = true;
  _isUnion = false;
  _isProduct = false;
  _nbDimensions = 5;
  _cardinal = M * ( E2+1 ) * ( E1+1 + N*( nu*E1+1 ) );
  // initializing zero state to (0,0,...,0)
  _zeroState = (int*)calloc( _nbDimensions, sizeof(int) );
  _zone = (marmoteSet**)NULL;
  _dimension = (marmoteSet**)NULL;

}

// destructor
layeredStateSpace::~layeredStateSpace()
{
  free( _zeroState );
}

// test for the "zero" state
bool layeredStateSpace::isZero(int *buffer)
{
  bool res = true;

  for ( int d = 0; res && ( d < _nbDimensions ); d++ ) {
    res = ( buffer[d] == 0 );
  }

  return res;
}

// initializing to the zero state
void layeredStateSpace::firstState(int *buffer)
{
  for ( int d = 0; d < _nbDimensions; d++ ) {
    buffer[d] = _zeroState[d];
  }
}

// incrementing the state
void layeredStateSpace::nextState(int *buffer)
{
  buffer[4]++;
  if ( buffer[4] == _M ) {
    buffer[3]++;
    buffer[4] = 0;
    if ( buffer[3] == _E2 + 1 ) {
      buffer[3] = 0;
      if ( buffer[0] == 0 ) {
        // We are in l(0). Just increment Q1
        buffer[2]++;
        if ( buffer[2] == _E1 + 1 ) {
          // jump to l(1,0)
          buffer[2] = 0;
          buffer[0] = 1;
        }
      }
      else if ( buffer[2] == 0 ) {
        // We are in some l(n,0). Jump to l-bar(n)
        buffer[1] = 1;
        buffer[2] = 1;
      }
      else {
        // We are in some l-bar(n). Increment Q1
        buffer[2]++;
        if ( buffer[2] == _E1 + 1 ) {
          buffer[1]++;
          buffer[2] = 1;
          if ( buffer[1] == _nu + 1 ) {
            // We finished l-bar(n). Jump to next level.
            buffer[0] ++;
             buffer[1] = 0;
            buffer[2] = 0;
            if ( buffer[0] == _N + 1 ) {
              // jump back to 0
              buffer[0] = 0;
            }
          }
        }
      }
    }
  }
}

// converting an index into an array
void layeredStateSpace::decodeState( int index, int* buf )
{
  int idx = index;

  if ( idx < _L0 ) {
    // this is l(0)
    buf[0] = 0;
    buf[1] = 0;
    buf[2] = idx / _S;
    idx = idx % _S;
    buf[3] = idx / _M;
    buf[4] = idx % _M;
  }
  else {
    idx -= _L0;
    buf[0] = 1 + idx / _L;
    idx = idx % _L;
    if ( idx < _S ) {
      // this is l(n,0)
      buf[1] = 0;
      buf[2] = 0;
      buf[3] = idx / _M;
      buf[4] = idx % _M;
    }
    else {
      idx -= _S;
      buf[1] = 1 + idx / ( _S * _E1 );
      idx = idx % ( _S * _E1 );
      buf[2] = 1 + idx / _S;
      idx = idx % _S;
      buf[3] = idx / _M;
      buf[4] = idx % _M;
    }
  }
}

// index of some state
int layeredStateSpace::index(int *buf)
{
  int res;

  if ( buf[0] == 0 ) {
    res = buf[2] * _S + buf[3] * _M + buf[4];
  }
  else {
    res = _L0 + ( buf[0] - 1 ) * _L;
    if ( buf[2] == 0 ) {
      res += buf[3] * _M + buf[4];
    }
    else {
      res += _S + ( buf[1] - 1 ) * _S * _E1 + ( buf[2] - 1 ) * _S + buf[3] * _M + buf[4];
    }
  }

  return res;
}

// printing
void layeredStateSpace::printState(FILE *out, int *buffer)
{
  fprintf( out, "( %3d; %3d, %3d, %3d, %3d )", buffer[0], buffer[1], buffer[2], buffer[3], buffer[4] );
}

// enumerating
void layeredStateSpace::enumerate()
{
  int stateBuffer[5];
  firstState(stateBuffer);

  int idx = 0;
  do {
    printState( stdout, stateBuffer );
    fprintf(stdout," idx: %5d (%d)", index(stateBuffer), idx );
    fprintf(stdout,"\n");
    nextState( stateBuffer );
    idx++;
  } while (!isZero(stateBuffer));

  for ( idx = 0; idx < cardinal(); idx++ ) {
    decodeState( idx, stateBuffer );
    printState( stdout, stateBuffer );
    fprintf(stdout," idx: %5d ", idx );
    fprintf(stdout,"\n");
  }
}
