

/**
 * @brief Procedure to initialize the problem data structure with the EPEW paper
 * values.
 * @param pbm a structure containing the data of the problem: contains initially
 * the principal values, and is filled by the procedure.
 */
void initData_EPEW_Paper( problemData *pbm ) {

  // strong specificity is that M = 2, nu = 2, K = 2.
  // The rest is more or less variable

  // Environment generator
  pbm->environmentTransition = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->environmentTransition[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->environmentTransition[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->environmentTransition[0], -0.01, 0.01 );
  INITARRAY( pbm->environmentTransition[1], 0.1, -0.1 );

  // Control thresholds. Add "guards" at 0 and K+1 but in effect,
  // threshold[K+1] may be smaller than threshold[K]
  pbm->threshold = (int*)malloc( (pbm->K+2)*sizeof(int) );
  pbm->threshold[0] = -1;
  pbm->threshold[1] = 8; pbm->threshold[2] = 12;
  pbm->threshold[pbm->K+1] = pbm->E1;

  // Table of data arrival rate: depends on environment phase and threshold interval
  pbm->dataRate = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->dataRate[0] = (double*)malloc( (pbm->K+1)*sizeof(double) );
  pbm->dataRate[1] = (double*)malloc( (pbm->K+1)*sizeof(double) );
  INITARRAY( pbm->dataRate[0], 1.5, 1.0 );
  INITARRAY( pbm->dataRate[1], 2.0, 1.5 );

  // Table of energy arrival rate: depends on EQ number and environment
  pbm->energyRate = (double**)malloc( 2*sizeof(double*) );
  pbm->energyRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->energyRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->energyRate[0], 1.5, 1.0 );
  INITARRAY( pbm->energyRate[1], 1.8, 1.2 );

  // Table of leak rates: depends on the EQ number and the environment
  pbm->leakRate = (double**)malloc( 2*sizeof(double*) );
  pbm->leakRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->leakRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->leakRate[0], 0.1, 0.3 );
  INITARRAY( pbm->leakRate[1], 0.3, 0.4 );

  // Table of signal generation rates: depend on threshold and environment
  pbm->signalRate = (double**)malloc( pbm->K*sizeof(double*) );
  pbm->signalRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->signalRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->signalRate[0], 0.3, 0.2 );
  INITARRAY( pbm->signalRate[1], 0.2, 0.1 );

  // Table of initial service phase probas: depend on the environment
  pbm->initServicePhaseProba = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->initServicePhaseProba[0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->initServicePhaseProba[1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->initServicePhaseProba[0], 0.2, 0.8 );
  INITARRAY( pbm->initServicePhaseProba[1], 0.7, 0.3 );

  // Service duration sub-generators: depend on the environment
  pbm->serviceTransition = (double***)malloc( pbm->M*sizeof(double**) );
  pbm->serviceTransition[0] = (double**)malloc( pbm->nu*sizeof(double*) );
  pbm->serviceTransition[0][0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->serviceTransition[0][1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->serviceTransition[0][0], -0.4368, 0.40768 );
  INITARRAY( pbm->serviceTransition[0][1], 0.42608, -1.71809 );
  pbm->serviceTransition[1] = (double**)malloc( pbm->nu*sizeof(double*) );
  pbm->serviceTransition[1][0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->serviceTransition[1][1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->serviceTransition[1][0], -0.6402, 0.56812 );
  INITARRAY( pbm->serviceTransition[1][1], 0.22308, -1.34325 );

  // Energy consumption probabilities: depend on the environment,
  // the service phase, and current energy level.
  // The formula used is generic wrt M and nu
  pbm->consumptionProba = (double****)malloc( pbm->M*sizeof(double***) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->consumptionProba[i] = (double***)malloc( pbm->nu*sizeof(double**) );
    for ( int x = 0; x < pbm->nu; x++ ) {
      pbm->consumptionProba[i][x] = (double**)calloc( (1+pbm->E1), sizeof(double*) );
      for ( int m1 = 1; m1 <= pbm->E1; m1++ ) {
        pbm->consumptionProba[i][x][m1] = (double*)calloc( (1+m1), sizeof(double) );
        double som = 0.0;
        for ( int k = 1; k <= m1; k++ ) {
          pbm->consumptionProba[i][x][m1][k] = pow( (1+i)*(1+x), -k );
          som += pbm->consumptionProba[i][x][m1][k];
        }
        for ( int k = 1; k <= m1; k++ ) {
          pbm->consumptionProba[i][x][m1][k] /= som;
        }
      }
    }
  }

  // Energy movement probabilities: depend on the environment,
  // the thresholds, and both current energy levels.
  // The formula used is generic wrt M and nu and does not depend on service phase.
  pbm->movementProba = (double*****)malloc( pbm->M*sizeof(double****) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->movementProba[i] = (double****)malloc( pbm->K*sizeof(double***) );
    for ( int s = 0; s < pbm->K; s++ ) {
      pbm->movementProba[i][s] = (double***)calloc( (1+pbm->E1), sizeof(double**) );
      for ( int m1 = 0; m1 <= pbm->E1; m1++ ) {
        pbm->movementProba[i][s][m1] = (double**)calloc( (1+pbm->E2), sizeof(double*) );
        for ( int m2 = 1; m2 <= pbm->E2; m2++ ) {
          pbm->movementProba[i][s][m1][m2] = (double*)calloc( (1+m2), sizeof(double) );
          double som = 0.0;
          for ( int k = 1; ( k <= m2 ) && ( k + m1 <= pbm->E1 ); k++ ) {
            pbm->movementProba[i][s][m1][m2][k] = pow( 2.0, -k*(i+1) );
            som += pbm->movementProba[i][s][m1][m2][k];
          }
          for ( int k = 1; ( k <= m2 ) && ( k + m1 <= pbm->E1 ); k++ ) {
            pbm->movementProba[i][s][m1][m2][k] /= som;
          }
        }
      }
    }
  }
}

/**
 * @brief Procedure to initialize the problem data structure with the M/M/1/K
 * values. Uses some structural parameters of the EPEW paper but considers all
 * environment states identical, all threshold values identical, and saturated energy queues.
 * Energy consumption and movement is left as in EPEW just to see how it works...
 * @param pbm a structure containing the data of the problem: contains initially
 * the principal values, and is filled by the procedure.
 */
void initData_MM1( problemData *pbm ) {

  // strong specificity is that M = 2, nu = 2, K = 2.
  // The rest is more or less variable

  // Environment generator
  pbm->environmentTransition = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->environmentTransition[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->environmentTransition[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->environmentTransition[0], -0.01, 0.01 );
  INITARRAY( pbm->environmentTransition[1], 0.1, -0.1 );

  // Control thresholds. Add "guards" at 0 and K+1 but in effect,
  // threshold[K+1] may be smaller than threshold[K]
  pbm->threshold = (int*)malloc( (pbm->K+2)*sizeof(int) );
  pbm->threshold[0] = -1;
  pbm->threshold[1] = 8; pbm->threshold[2] = 10;
  pbm->threshold[pbm->K+1] = pbm->E1;

  // Table of data arrival rate: depends on environment phase and threshold interval
  pbm->dataRate = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->dataRate[0] = (double*)malloc( pbm->K*sizeof(double) );
  pbm->dataRate[1] = (double*)malloc( pbm->K*sizeof(double) );
  INITARRAY( pbm->dataRate[0], 1.0, 1.0 );
  INITARRAY( pbm->dataRate[1], 1.0, 1.0 );

  // Table of energy arrival rate: depends on EQ number and environment
  pbm->energyRate = (double**)malloc( 2*sizeof(double*) );
  pbm->energyRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->energyRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->energyRate[0], 50.0, 50.0 );
  INITARRAY( pbm->energyRate[1], 50.0, 50.0 );

  // Table of leak rates: depends on the EQ number and the environment
  pbm->leakRate = (double**)malloc( 2*sizeof(double*) );
  pbm->leakRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->leakRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->leakRate[0], 0.0, 0.0 );
  INITARRAY( pbm->leakRate[1], 0.0, 0.0 );

  // Table of signal generation rates: depend on threshold and environment
  pbm->signalRate = (double**)malloc( pbm->K*sizeof(double*) );
  pbm->signalRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->signalRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->signalRate[0], 30.0, 20.0 );
  INITARRAY( pbm->signalRate[1], 25.0, 45.0 );

  // Table of initial service phase probas: depend on the environment
  pbm->initServicePhaseProba = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->initServicePhaseProba[0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->initServicePhaseProba[1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->initServicePhaseProba[0], 0.2, 0.8 );
  INITARRAY( pbm->initServicePhaseProba[1], 0.7, 0.3 );

  // Service duration sub-generators: depend on the environment
  pbm->serviceTransition = (double***)malloc( pbm->M*sizeof(double**) );
  pbm->serviceTransition[0] = (double**)malloc( pbm->nu*sizeof(double*) );
  pbm->serviceTransition[0][0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->serviceTransition[0][1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->serviceTransition[0][0], -2.0, 0.0 );
  INITARRAY( pbm->serviceTransition[0][1], 0.0, -2.0 );
  pbm->serviceTransition[1] = (double**)malloc( pbm->nu*sizeof(double*) );
  pbm->serviceTransition[1][0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->serviceTransition[1][1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->serviceTransition[1][0], -2.0, 0.0 );
  INITARRAY( pbm->serviceTransition[1][1], 0.0, -2.0 );

  // Energy consumption probabilities: depend on the environment,
  // the service phase, and current energy level.
  // The formula used is generic wrt M and nu
  pbm->consumptionProba = (double****)malloc( pbm->M*sizeof(double***) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->consumptionProba[i] = (double***)malloc( pbm->nu*sizeof(double**) );
    for ( int x = 0; x < pbm->nu; x++ ) {
      pbm->consumptionProba[i][x] = (double**)calloc( (1+pbm->E1), sizeof(double*) );
      for ( int m1 = 1; m1 <= pbm->E1; m1++ ) {
        pbm->consumptionProba[i][x][m1] = (double*)calloc( (1+m1), sizeof(double) );
        double som = 0.0;
        for ( int k = 1; k <= m1; k++ ) {
          pbm->consumptionProba[i][x][m1][k] = pow( (1+i)*(1+x), -k );
          som += pbm->consumptionProba[i][x][m1][k];
        }
        for ( int k = 1; k <= m1; k++ ) {
          pbm->consumptionProba[i][x][m1][k] /= som;
        }
      }
    }
  }

  // Energy movement probabilities: depend on the environment,
  // the thresholds, and both current energy levels.
  // The formula used is generic wrt M and nu and does not depend on service phase.
  pbm->movementProba = (double*****)malloc( pbm->M*sizeof(double****) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->movementProba[i] = (double****)malloc( pbm->K*sizeof(double***) );
    for ( int s = 0; s < pbm->K; s++ ) {
      pbm->movementProba[i][s] = (double***)calloc( (1+pbm->E1), sizeof(double**) );
      for ( int m1 = 0; m1 <= pbm->E1; m1++ ) {
        pbm->movementProba[i][s][m1] = (double**)calloc( (1+pbm->E2), sizeof(double*) );
        for ( int m2 = 1; m2 <= pbm->E2; m2++ ) {
          pbm->movementProba[i][s][m1][m2] = (double*)calloc( (1+m2), sizeof(double) );
          double som = 0.0;
          for ( int k = 1; ( k <= m2 ) && ( k + m1 <= pbm->E1 ); k++ ) {
            pbm->movementProba[i][s][m1][m2][k] = pow( 2.0, -k*(i+1) );
            som += pbm->movementProba[i][s][m1][m2][k];
          }
          for ( int k = 1; ( k <= m2 ) && ( k + m1 <= pbm->E1 ); k++ ) {
            pbm->movementProba[i][s][m1][m2][k] /= som;
          }
        }
      }
    }
  }
}

/**
 * @brief Procedure to initialize the problem data structure with values for a large
 * experiment. Uses a rather fine grain consumption model. Thresholds are evenly spaced
 * and traffic control is linear.
 * @param pbm a structure containing the data of the problem: contains initially
 * the principal values, and is filled by the procedure.
 */
void initData_bigBatteries( problemData *pbm ) {

  // strong specificity is that M = 2, nu = 2
  // The rest is more or less variable

  // Environment generator
  pbm->environmentTransition = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->environmentTransition[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->environmentTransition[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->environmentTransition[0], -0.01, 0.01 );
  INITARRAY( pbm->environmentTransition[1], 0.1, -0.1 );

  // Control thresholds. Add "guards" at 0 and K+1 but in effect,
  // threshold[K+1] may be smaller than threshold[K]
  pbm->threshold = (int*)malloc( (pbm->K+2)*sizeof(int) );
  pbm->threshold[0] = -1;
  pbm->threshold[pbm->K+1] = pbm->E1;
  for ( int k = 1; k <= pbm->K; k++ ) {
    pbm->threshold[k] = (pbm->E1*k)/(pbm->K+1);
  }

  // Table of data arrival rate: depends on environment phase and threshold interval
  pbm->dataRate = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->dataRate[0] = (double*)malloc( pbm->K*sizeof(double) );
  pbm->dataRate[1] = (double*)malloc( pbm->K*sizeof(double) );
  for ( int k = 0; k < pbm->K; k++ ) {
    pbm->dataRate[0][k] =  1.5 * (pbm->K-k) / (double)pbm->K;
    pbm->dataRate[0][k] =  2.0 * (pbm->K-k) / (double)pbm->K;
  }

  // Table of energy arrival rate: depends on EQ number and environment
  pbm->energyRate = (double**)malloc( 2*sizeof(double*) );
  pbm->energyRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->energyRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->energyRate[0], 50.0, 50.0 );
  INITARRAY( pbm->energyRate[1], 50.0, 50.0 );

  // Table of leak rates: depends on the EQ number and the environment
  pbm->leakRate = (double**)malloc( 2*sizeof(double*) );
  pbm->leakRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->leakRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  INITARRAY( pbm->leakRate[0], 0.1, 0.3 );
  INITARRAY( pbm->leakRate[1], 0.3, 0.4 );

  // Table of signal generation rates: depend on threshold and environment
  pbm->signalRate = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->signalRate[0] = (double*)malloc( pbm->K*sizeof(double) );
  pbm->signalRate[1] = (double*)malloc( pbm->K*sizeof(double) );
  for ( int k = 0; k < pbm->K; k++ ) {
    pbm->signalRate[0][k] =  15 * (k+1) / (double)pbm->K;
    pbm->signalRate[1][k] =  20 * (k+1) / (double)pbm->K;
  }

  // Table of initial service phase probas: depend on the environment
  pbm->initServicePhaseProba = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->initServicePhaseProba[0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->initServicePhaseProba[1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->initServicePhaseProba[0], 0.2, 0.8 );
  INITARRAY( pbm->initServicePhaseProba[1], 0.7, 0.3 );

  // Service duration sub-generators: depend on the environment
  pbm->serviceTransition = (double***)malloc( pbm->M*sizeof(double**) );
  pbm->serviceTransition[0] = (double**)malloc( pbm->nu*sizeof(double*) );
  pbm->serviceTransition[0][0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->serviceTransition[0][1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->serviceTransition[0][0], -2.0, 0.0 );
  INITARRAY( pbm->serviceTransition[0][1], 0.0, -2.0 );
  pbm->serviceTransition[1] = (double**)malloc( pbm->nu*sizeof(double*) );
  pbm->serviceTransition[1][0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->serviceTransition[1][1] = (double*)malloc( pbm->nu*sizeof(double) );
  INITARRAY( pbm->serviceTransition[1][0], -2.0, 0.0 );
  INITARRAY( pbm->serviceTransition[1][1], 0.0, -2.0 );

  // Energy consumption probabilities: chosen independent on the environment
  // and the phase and almost from the energy level.
  // The formula used is generic wrt M and nu
  pbm->consumptionProba = (double****)malloc( pbm->M*sizeof(double***) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->consumptionProba[i] = (double***)malloc( pbm->nu*sizeof(double**) );
    for ( int x = 0; x < pbm->nu; x++ ) {
      pbm->consumptionProba[i][x] = (double**)calloc( (1+pbm->E1), sizeof(double*) );
      for ( int m1 = 1; m1 <= pbm->E1; m1++ ) {
        pbm->consumptionProba[i][x][m1] = (double*)calloc( (1+m1), sizeof(double) );
        double som = 0.0;
        for ( int k = 0; ( k <= 2 ) && ( k <= m1 ); k++ ) {
          pbm->consumptionProba[i][x][m1][k] = 1.0;
          som += pbm->consumptionProba[i][x][m1][k];
        }
        for ( int k = 1; ( k <= 2 ) && ( k <= m1 ); k++ ) {
          pbm->consumptionProba[i][x][m1][k] /= som;
        }
      }
    }
  }

  // Energy movement probabilities: depend on the environment,
  // the thresholds, and both current energy levels.
  // The formula used is generic wrt M and nu and does not depend on service phase.
  pbm->movementProba = (double*****)malloc( pbm->M*sizeof(double****) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->movementProba[i] = (double****)malloc( pbm->K*sizeof(double***) );
    for ( int s = 0; s < pbm->K; s++ ) {
      pbm->movementProba[i][s] = (double***)calloc( (1+pbm->E1), sizeof(double**) );
      for ( int m1 = 0; m1 <= pbm->E1; m1++ ) {
        pbm->movementProba[i][s][m1] = (double**)calloc( (1+pbm->E2), sizeof(double*) );
        for ( int m2 = 1; m2 <= pbm->E2; m2++ ) {
          pbm->movementProba[i][s][m1][m2] = (double*)calloc( (1+m2), sizeof(double) );
          double som = 0.0;
          for ( int k = 1; ( k <= m2 ) && ( k + m1 <= pbm->E1 ); k++ ) {
            pbm->movementProba[i][s][m1][m2][k] = pow( 2.0, -k*(i+1) );
            som += pbm->movementProba[i][s][m1][m2][k];
          }
          for ( int k = 1; ( k <= m2 ) && ( k + m1 <= pbm->E1 ); k++ ) {
            pbm->movementProba[i][s][m1][m2][k] /= som;
          }
        }
      }
    }
  }
}

/**
 * @brief Procedure to initialize the problem data structure with values for the
 * experiment close to the product-form model.
 * @param pbm a structure containing the data of the problem: contains initially
 * the principal values, and is filled by the procedure.
 */
void initData_productForm( problemData *pbm ) {

  // strong specificity is that M = 1, nu = 1

  // Environment generator. Only one state.
  pbm->environmentTransition = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->environmentTransition[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->environmentTransition[0][0] = 0.0;

  // Control thresholds. Only one zone (K=0)
  pbm->threshold = (int*)malloc( (pbm->K+2)*sizeof(int) );
  pbm->threshold[0] = -1;
  pbm->threshold[pbm->K+1] = pbm->E1;

  // Table of data arrival rate: depends on environment phase and threshold interval
  pbm->dataRate = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->dataRate[0] = (double*)malloc( (pbm->K+1)*sizeof(double) );
  pbm->dataRate[0][0] = pbm->lambda_p;

  // Table of energy arrival rate: depends on EQ number and environment
  pbm->energyRate = (double**)malloc( 2*sizeof(double*) );
  pbm->energyRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->energyRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->energyRate[0][0] = pbm->lambda_e1;
  pbm->energyRate[1][0] = pbm->lambda_e2;

  // Table of leak rates: depends on the EQ number and the environment
  pbm->leakRate = (double**)malloc( 2*sizeof(double*) );
  pbm->leakRate[0] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->leakRate[1] = (double*)malloc( pbm->M*sizeof(double) );
  pbm->leakRate[0][0] = pbm->u1;
  pbm->leakRate[1][0] = pbm->u2;

  // Table of signal generation rates: constant
  pbm->signalRate = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->signalRate[0] = (double*)malloc( (pbm->K+1)*sizeof(double) );
  pbm->signalRate[0][0] = pbm->lambda_minus;

  // Table of initial service phase probas: only one phase
  pbm->initServicePhaseProba = (double**)malloc( pbm->M*sizeof(double*) );
  pbm->initServicePhaseProba[0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->initServicePhaseProba[0][0] = 1.0;

  // Service duration sub-generator: exponential with very small value
  pbm->serviceTransition = (double***)malloc( pbm->M*sizeof(double**) );
  pbm->serviceTransition[0] = (double**)malloc( pbm->nu*sizeof(double*) );
  pbm->serviceTransition[0][0] = (double*)malloc( pbm->nu*sizeof(double) );
  pbm->serviceTransition[0][0][0] = -100.0;

  // Energy consumption probabilities: chosen independent on the environment
  // and the phase and almost from the energy level.
  // The formula used is generic wrt M and nu
  pbm->consumptionProba = (double****)malloc( pbm->M*sizeof(double***) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->consumptionProba[i] = (double***)malloc( pbm->nu*sizeof(double**) );
    for ( int x = 0; x < pbm->nu; x++ ) {
      pbm->consumptionProba[i][x] = (double**)calloc( (1+pbm->E1), sizeof(double*) );
      for ( int m1 = 1; m1 <= pbm->E1; m1++ ) {
        pbm->consumptionProba[i][x][m1] = (double*)calloc( (1+m1), sizeof(double) );
        pbm->consumptionProba[i][x][m1][1] = 1.0;
      }
    }
  }

  // Energy movement probabilities: depend on the environment,
  // the thresholds, and both current energy levels.
  // The formula used is generic wrt M and nu and does not depend on service phase.
  pbm->movementProba = (double*****)malloc( pbm->M*sizeof(double****) );
  for ( int i = 0; i < pbm->M; i++ ) {
    pbm->movementProba[i] = (double****)malloc( (pbm->K+1)*sizeof(double***) );
    for ( int s = 0; s <= pbm->K; s++ ) {
      pbm->movementProba[i][s] = (double***)calloc( (1+pbm->E1), sizeof(double**) );
      for ( int m1 = 0; m1 <= pbm->E1; m1++ ) {
        pbm->movementProba[i][s][m1] = (double**)calloc( (1+pbm->E2), sizeof(double*) );
        for ( int m2 = 1; m2 <= pbm->E2; m2++ ) {
          pbm->movementProba[i][s][m1][m2] = (double*)calloc( (1+m2), sizeof(double) );
          pbm->movementProba[i][s][m1][m2][1] = 1.0;
        }
      }
    }
  }
}
