#include "layeredStateSpace.h"
#include "markovChain.h"
#include "transitionStructure/sparseMatrix.h"
#include <iostream>
#include <stdlib.h>

#define INITARRAY1( a, b ) a[0] = b[0]; a[1] = b[1];
#define INITARRAY( a, b, c ) a[0] = b; a[1] = c;

struct problemData {
  int N;
  int K;
  int E1;
  int E2;
  int M;
  int nu;
  double **environmentTransition;
  int *threshold;
  double **dataRate;
  double **energyRate;
  double **leakRate;
  double **signalRate;
  double **initServicePhaseProba;
  double*** serviceTransition;
  double ****consumptionProba;
  double *****movementProba;
  double lambda_p;
  double lambda_e1;
  double lambda_e2;
  double u1;
  double u2;
  double lambda_minus;
};

#include "models.cpp"
#include "solutions.cpp"

void deleteData( problemData* pbm ) {

  free( pbm->threshold );

  for ( int i = 0; i < pbm->M; i++ ) {
    free( pbm->environmentTransition[i] );
  }
  free( pbm->environmentTransition );

  for ( int k = 0; k < pbm->K; k++ ) {
    free( pbm->signalRate[k] );
  }
  free( pbm->signalRate );

  for ( int i = 0; i < 2; i++ ) {
    free( pbm->energyRate[i] );
    free( pbm->leakRate[i] );
  }
  free( pbm->energyRate );
  free( pbm->leakRate );

  for ( int i = 0; i < pbm->M; i++ ) {
    free( pbm->dataRate[i] );
    free( pbm->initServicePhaseProba[i] );
  }
  free( pbm->dataRate );
  free( pbm->initServicePhaseProba );

  for ( int i = 0; i < pbm->M; i++ ) {
    for ( int j = 0; j < pbm->nu; j++ ) {
      free( pbm->serviceTransition[i][j] );
    }
    free( pbm->serviceTransition[i] );
  }
  free( pbm->serviceTransition );

  for ( int i = 0; i < pbm->M; i++ ) {
    for ( int x = 0; x < pbm->nu; x++ ) {
      for ( int m1 = 1; m1 <= pbm->E1; m1++ ) {
        free( pbm->consumptionProba[i][x][m1] );
      }
      free( pbm->consumptionProba[i][x] );
    }
    free( pbm->consumptionProba[i] );
  }
  free( pbm->consumptionProba );

  for ( int i = 0; i < pbm->M; i++ ) {
    for ( int s = 0; s < pbm->K; s++ ) {
      for ( int m1 = 0; m1 <= pbm->E1; m1++ ) {
        for ( int m2 = 1; m2 <= pbm->E2; m2++ ) {
          free( pbm->movementProba[i][s][m1][m2] );
        }
        free( pbm->movementProba[i][s][m1] );
      }
      free( pbm->movementProba[i][s] );
    }
    free( pbm->movementProba[i] );
  }
  free( pbm->movementProba );

}

sparseMatrix* makeGenerator(layeredStateSpace* sp, problemData *pbm ) {

  // naming of coordinates
  int DATA = 0;
  int SERV = 1;
  int DEQ  = 2;
  int SEQ  = 3;
  int RE   = 4;

  // mapping of energy values in the DEQ to control zones.
  // Note that threshold[0] = 0 and threshold[K+1] = E1
  int *controlMap = (int*)calloc( (1+pbm->E1), sizeof(int) );
  for ( int s = 0; s <= pbm->K; s++ ) {
    for ( int m1 = pbm->threshold[s]+1; ( m1 <= pbm->threshold[s+1] ) && ( m1 <= pbm->E1 ); m1++ ) {
      controlMap[m1] = s;
    }
  }

  // creation of the (empty) transition object
  sparseMatrix* gen = new sparseMatrix( sp->cardinal() );

  int stateBuffer[5];
  sp->firstState(stateBuffer);

  int idx = 0;
  do {
    // std::cout << "Doing state #" << idx << ": ";
    // sp->printState(stdout,stateBuffer);
    // std::cout << " (idx = " << sp->index(stateBuffer) << ")";
    // std::cout << std::endl;

    // convenience local variables, also used for restoring stateBuffer
    int m1 = stateBuffer[DEQ];
    int m2 = stateBuffer[SEQ];
    int env = stateBuffer[RE];
    int sphase = stateBuffer[SERV];
    int control = controlMap[m1];

    // environment change
    for ( int i = 0; i < pbm->M; i++ ) {
      if ( i != env ) {
        stateBuffer[RE] = i;
        gen->addToEntry( idx, sp->index(stateBuffer), pbm->environmentTransition[env][i] );
        gen->addToEntry( idx, idx, -pbm->environmentTransition[env][i] );
      }
    }
    stateBuffer[RE] = env;

    // packet arrival
    if ( stateBuffer[DATA] < pbm->N ) {
      stateBuffer[DATA]++;
      if ( stateBuffer[DATA] == 1 ) {
        // packet just arrived in an empty queue
        if ( stateBuffer[DEQ] > 0 ) {
          // start service
          for ( int x = 0; x < pbm->nu; x++ ) {
            // start it in phase x
            stateBuffer[SERV] = x+1;
            gen->addToEntry( idx, sp->index(stateBuffer), pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
            gen->addToEntry( idx, idx, -pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
          }
          stateBuffer[SERV] = 0;
        }
        else {
          // just queue packet
          gen->addToEntry( idx, sp->index(stateBuffer), pbm->dataRate[env][control] );
          gen->addToEntry( idx, idx, -pbm->dataRate[env][control] );
        }
      }
      else {
        // just queue packet
        gen->addToEntry( idx, sp->index(stateBuffer), pbm->dataRate[env][control] );
        gen->addToEntry( idx, idx, -pbm->dataRate[env][control] );
      }
      stateBuffer[DATA]--;
    }

    // energy arrival
    if ( stateBuffer[DEQ] == 0 ) {
      // case where the DEQ was empty: possibility of starting a service
      if ( stateBuffer[DATA] > 0 ) {
        // start service
        for ( int x = 0; x < pbm->nu; x++ ) {
          // start it in phase x
          stateBuffer[SERV] = x+1;
          gen->addToEntry( idx, sp->index(stateBuffer), pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
          gen->addToEntry( idx, idx, -pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
        }
        stateBuffer[SERV] = 0;
      }
      stateBuffer[DEQ] = 0;
    }
    else if ( stateBuffer[DEQ] < pbm->E1 ) {
      stateBuffer[DEQ]++;
      gen->addToEntry( idx, sp->index(stateBuffer), pbm->energyRate[0][env] );
      gen->addToEntry( idx, idx, -pbm->energyRate[0][env] );
      stateBuffer[DEQ]--;
    }

    if ( stateBuffer[SEQ] < pbm->E2 ) {
      stateBuffer[SEQ]++;
      gen->addToEntry( idx, sp->index(stateBuffer), pbm->energyRate[1][env] );
      gen->addToEntry( idx, idx, -pbm->energyRate[1][env] );
      stateBuffer[SEQ]--;
    }

    // energy leakage
    if ( stateBuffer[DEQ] > 0 ) {
      stateBuffer[DEQ]--;
      if ( stateBuffer[DEQ] == 0 ) {
        // energy shortage. Cancel current service. TO BE DONE
        gen->addToEntry( idx, sp->index(stateBuffer), pbm->leakRate[0][env] );
        gen->addToEntry( idx, idx, -pbm->leakRate[0][env] );
      }
      else {
        gen->addToEntry( idx, sp->index(stateBuffer), pbm->leakRate[0][env] );
        gen->addToEntry( idx, idx, -pbm->leakRate[0][env] );
      }
      stateBuffer[DEQ]++;
    }

    if ( stateBuffer[SEQ] > 0 ) {
      stateBuffer[SEQ]--;
      gen->addToEntry( idx, sp->index(stateBuffer), pbm->leakRate[1][env] );
      gen->addToEntry( idx, idx, -pbm->leakRate[1][env] );
      stateBuffer[SEQ]++;
    }

    // service phase/packet departure
    if ( stateBuffer[SERV] > 0 ) {
      // assert: sphase > 0

      double endServiceRate = -pbm->serviceTransition[env][sphase-1][sphase-1];
      if ( endServiceRate > 0.0 ) {
        // just in case data would prescribe an infinite service time...

        // phase transition implies at the same time energy consumption and new phase:
        // this is why there is a double loop.

        // Outer loop: phase completion implies energy consumption, between 1 and m1
        for ( int k = 1; k <= m1; k++ ) {
          stateBuffer[DEQ] -= k;
          double proba = pbm->consumptionProba[env][sphase-1][m1][k];

          for ( int y = 0; y < pbm->nu; y++ ) {
            if ( y != sphase-1 ) {
              // transition from sphase to y without service completion
              stateBuffer[SERV] = y;
              gen->addToEntry( idx, sp->index(stateBuffer), proba*pbm->serviceTransition[env][sphase-1][y] );
              gen->addToEntry( idx, idx, -proba*pbm->serviceTransition[env][sphase-1][y] );
              endServiceRate -= pbm->serviceTransition[env][sphase-1][y];
            }
          }
          // Service completion, with the remaining rate endServiceRate. It may be that
          // this rate is 0 (just phase transitions): we have to test again.
          stateBuffer[DATA]--;
          if ( endServiceRate > 0.0 ) {
            // Ok there is a possibility of ending the service. Does it imply beginning of service?
            if ( ( stateBuffer[DATA] > 0 ) && ( stateBuffer[DEQ] > 0 ) ) {
              // Yes: there is enough data AND energy to start service.
              // Loop on possible start phases.
              for ( int x = 0; x < pbm->nu; x++ ) {
                // start it in phase x
                stateBuffer[SERV] = x+1;
                gen->addToEntry( idx, sp->index(stateBuffer), proba*endServiceRate*pbm->initServicePhaseProba[env][x] );
                gen->addToEntry( idx, idx, -proba*endServiceRate*pbm->initServicePhaseProba[env][x] );
              }
              stateBuffer[SERV] = 0;
            }
            else {
              // No service starts: just jump down, consume energy and stop server
              stateBuffer[SERV] = 0;
              gen->addToEntry( idx, sp->index(stateBuffer), proba*endServiceRate );
              gen->addToEntry( idx, idx, -proba*endServiceRate );
            }
          }
          stateBuffer[DATA]++;
          stateBuffer[DEQ] += k;
        }
        stateBuffer[SERV] = sphase;
      }
    }

    // signals
    if ( ( stateBuffer[SEQ] > 0 ) && ( stateBuffer[DEQ] < pbm->E1 ) ){
      double rate = pbm->signalRate[env][control];
      if ( rate > 0.0 ) {

        for ( int k = 1; ( k <= m2 ) && ( m1 + k <= pbm->E1 ); k++ ) {
          stateBuffer[SEQ] -= k;
          stateBuffer[DEQ] += k;

          // there is the possibility that service starts
          if ( stateBuffer[SERV] == 0 ) {
            for ( int x = 0; x < pbm->nu; x++ ) {
              // start it in phase x
              stateBuffer[SERV] = x+1;
              double taux = rate*pbm->movementProba[env][control][m1][m2][k]*pbm->initServicePhaseProba[env][x];
              gen->addToEntry( idx, sp->index(stateBuffer), taux );
              gen->addToEntry( idx, idx, -taux );
            }
            stateBuffer[SERV] = 0;
          }
          // or just that energy moves
          else {
            gen->addToEntry( idx, sp->index(stateBuffer), rate*pbm->movementProba[env][control][m1][m2][k] );
            gen->addToEntry( idx, idx, -rate*pbm->movementProba[env][control][m1][m2][k] );
          }

          stateBuffer[DEQ] -= k;
          stateBuffer[SEQ] += k;
        }
      }
    }

    sp->nextState( stateBuffer );
    idx++;
  } while (!sp->isZero(stateBuffer));

  gen->setType( CONTINUOUS );

  // cleanup
  free( controlMap );

  // done!
  return gen;
}

sparseMatrix** makeGeneratorElements(layeredStateSpace* sp, problemData *pbm ) {

  // naming of coordinates
  int DATA = 0;
  int SERV = 1;
  int DEQ  = 2;
  int SEQ  = 3;
  int RE   = 4;
  // naming of elements
  int B0     = 0;
  int Ctilde = 1;
  int A10    = 2;
  int A1     = 3;
  int C      = 4;
  int A21    = 5;
  int A2     = 6;

  // mapping of energy values in the DEQ to control zones.
  // Note that threshold[0] = 0 and threshold[K+1] = E1
  int *controlMap = (int*)calloc( (1+pbm->E1), sizeof(int) );
  for ( int s = 0; s <= pbm->K; s++ ) {
    for ( int m1 = pbm->threshold[s]+1; ( m1 <= pbm->threshold[s+1] ) && ( m1 <= pbm->E1 ); m1++ ) {
      controlMap[m1] = s;
    }
  }

  // creation of the (empty) transition objects
  int L0 = pbm->M * ( 1+pbm->E1 ) * ( 1+pbm->E2 );
  int L = pbm->M * ( 1+pbm->nu*pbm->E1 ) * ( 1+pbm->E2 );
  sparseMatrix** res = (sparseMatrix**) malloc( 7*sizeof(sparseMatrix*) );
  res[B0] = new sparseMatrix( L0 );
  res[Ctilde] = new sparseMatrix( L0, L ); // !! not square
  res[A10] = new sparseMatrix( L, L0 ); // !! not square
  res[A1] = new sparseMatrix( L );
  res[C] = new sparseMatrix( L );
  res[A21] = new sparseMatrix( L );
  res[A2] = new sparseMatrix( L );

  int stateBuffer[5];
  sp->firstState(stateBuffer);

  int idx = 0;
  do {
    // std::cout << "Doing state #" << idx << ": ";
    // sp->printState(stdout,stateBuffer);
    // std::cout << " (idx = " << sp->index(stateBuffer) << ")";
    // std::cout << std::endl;

    // convenience local variables, also used for restoring stateBuffer
    int m1 = stateBuffer[DEQ];
    int m2 = stateBuffer[SEQ];
    int env = stateBuffer[RE];
    int sphase = stateBuffer[SERV];
    int control = controlMap[m1];

    // must be calculated before changing DATA
    int WHERE = ( stateBuffer[DATA] == 0 ? B0 : ( stateBuffer[DATA] == pbm->N ? A2 : A1 ) );
    int DOWN = ( stateBuffer[DATA] == 1 ? A10 : A21 );
#define IDX(n) ( n < L0 ? n : ( n - L0 ) % L )

    // Matrices are different only on rows 0, 1 and N.
    if ( ( stateBuffer[DATA] > 1 ) && ( stateBuffer[DATA] < pbm->N ) ) {
      sp->nextState( stateBuffer );
      idx++;
      continue;
    }

    // environment change
    for ( int i = 0; i < pbm->M; i++ ) {
      if ( i != env ) {
        stateBuffer[RE] = i;
        res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->environmentTransition[env][i] );
        res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->environmentTransition[env][i] );
      }
    }
    stateBuffer[RE] = env;


    // packet arrival
    if ( stateBuffer[DATA] < pbm->N ) {
      stateBuffer[DATA]++;
      if ( stateBuffer[DATA] == 1 ) {
        // packet just arrived in an empty queue
        if ( stateBuffer[DEQ] > 0 ) {
          // start service
          for ( int x = 0; x < pbm->nu; x++ ) {
            // start it in phase x
            stateBuffer[SERV] = x+1;
            res[Ctilde]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
            res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
          }
          stateBuffer[SERV] = 0;
        }
        else {
          // just queue packet
          res[Ctilde]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->dataRate[env][control] );
          res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->dataRate[env][control] );
        }
      }
      else {
        // just queue packet
        // cout << "Putting value " << pbm->dataRate[env][control] << " into C[" << IDX(idx) << "," << IDX(sp->index(stateBuffer)) << "]" << endl;
        res[C]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->dataRate[env][control] );
        res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->dataRate[env][control] );
      }
      stateBuffer[DATA]--;
    }

    // energy arrival
    if ( stateBuffer[DEQ] == 0 ) {
      // case where the DEQ was empty: possibility of starting a service
      if ( stateBuffer[DATA] > 0 ) {
        // start service
        for ( int x = 0; x < pbm->nu; x++ ) {
          // start it in phase x
          stateBuffer[SERV] = x+1;
          res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
          res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->dataRate[env][control]*pbm->initServicePhaseProba[env][x] );
        }
        stateBuffer[SERV] = 0;
      }
      stateBuffer[DEQ] = 0;
    }
    else if ( stateBuffer[DEQ] < pbm->E1 ) {
      stateBuffer[DEQ]++;
      res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->energyRate[0][env] );
      res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->energyRate[0][env] );
      stateBuffer[DEQ]--;
    }

    if ( stateBuffer[SEQ] < pbm->E2 ) {
      stateBuffer[SEQ]++;
      res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->energyRate[1][env] );
      res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->energyRate[1][env] );
      stateBuffer[SEQ]--;
    }

    // energy leakage
    if ( stateBuffer[DEQ] > 0 ) {
      stateBuffer[DEQ]--;
      res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->leakRate[0][env] );
      res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->leakRate[0][env] );
      stateBuffer[DEQ]++;
    }

    if ( stateBuffer[SEQ] > 0 ) {
      stateBuffer[SEQ]--;
      res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), pbm->leakRate[1][env] );
      res[WHERE]->addToEntry( IDX(idx), IDX(idx), -pbm->leakRate[1][env] );
      stateBuffer[SEQ]++;
    }

    // service phase/packet departure
    if ( stateBuffer[SERV] > 0 ) {
      // assert: sphase > 0

      double endServiceRate = -pbm->serviceTransition[env][sphase-1][sphase-1];
      if ( endServiceRate > 0.0 ) {
        // just in case data would prescribe an infinite service time...

        // phase transition implies at the same time energy consumption and new phase:
        // this is why there is a double loop.

        // Outer loop: phase completion implies energy consumption, between 1 and m1
        for ( int k = 1; k <= m1; k++ ) {
          stateBuffer[DEQ] -= k;
          double proba = pbm->consumptionProba[env][sphase-1][m1][k];

          for ( int y = 0; y < pbm->nu; y++ ) {
            if ( y != sphase-1 ) {
              // transition from sphase to y without service completion
              stateBuffer[SERV] = y;
              res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), proba*pbm->serviceTransition[env][sphase-1][y] );
              res[WHERE]->addToEntry( IDX(idx), IDX(idx), -proba*pbm->serviceTransition[env][sphase-1][y] );
              endServiceRate -= pbm->serviceTransition[env][sphase-1][y];
            }
          }
          // Service completion, with the remaining rate endServiceRate. It may be that
          // this rate is 0 (just phase transitions): we have to test again.
          stateBuffer[DATA]--;
          if ( endServiceRate > 0.0 ) {
            // Ok there is a possibility of ending the service. Does it imply beginning of service?
            if ( ( stateBuffer[DATA] > 0 ) && ( stateBuffer[DEQ] > 0 ) ) {
              // Yes: there is enough data AND energy to start service.
              // Loop on possible start phases.
              for ( int x = 0; x < pbm->nu; x++ ) {
                // start it in phase x
                stateBuffer[SERV] = x+1;
                res[DOWN]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), proba*endServiceRate*pbm->initServicePhaseProba[env][x] );
                res[WHERE]->addToEntry( IDX(idx), IDX(idx), -proba*endServiceRate*pbm->initServicePhaseProba[env][x] );
              }
              stateBuffer[SERV] = 0;
            }
            else {
              // No service starts: just jump down, consume energy and stop server
              stateBuffer[SERV] = 0;
              res[DOWN]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), proba*endServiceRate );
              res[WHERE]->addToEntry( IDX(idx), IDX(idx), -proba*endServiceRate );
            }
          }
          stateBuffer[DATA]++;
          stateBuffer[DEQ] += k;
        }
        stateBuffer[SERV] = sphase;
      }
    }

    // signals
    if ( ( stateBuffer[SEQ] > 0 ) && ( stateBuffer[DEQ] < pbm->E1 ) ){
      double rate = pbm->signalRate[control][env];
      if ( rate > 0.0 ) {

        for ( int k = 1; ( k <= m2 ) && ( m1 + k <= pbm->E1 ); k++ ) {
          stateBuffer[SEQ] -= k;
          stateBuffer[DEQ] += k;

          // there is the possibility that service starts
          if ( stateBuffer[SERV] == 0 ) {
            for ( int x = 0; x < pbm->nu; x++ ) {
              // start it in phase x
              stateBuffer[SERV] = x+1;
              double taux = rate*pbm->movementProba[env][control][m1][m2][k]*pbm->initServicePhaseProba[env][x];
              res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), taux );
              res[WHERE]->addToEntry( IDX(idx), IDX(idx), -taux );
            }
            stateBuffer[SERV] = 0;
          }
          // or just that energy moves
          else {
            res[WHERE]->addToEntry( IDX(idx), IDX(sp->index(stateBuffer)), rate*pbm->movementProba[env][control][m1][m2][k] );
            res[WHERE]->addToEntry( IDX(idx), IDX(idx), -rate*pbm->movementProba[env][control][m1][m2][k] );
          }

          stateBuffer[DEQ] -= k;
          stateBuffer[SEQ] += k;
        }
      }
    }

    sp->nextState( stateBuffer );
    idx++;
  } while (!sp->isZero(stateBuffer));

  // cleanup
  free( controlMap );

  // done!
  return res;
}

void compareRepresentations( problemData* pbm, layeredStateSpace* sp, sparseMatrix* gen, sparseMatrix** block)
{
  // naming of coordinates
  int DATA = 0;
  //int SERV = 1;
  //int DEQ  = 2;
  //int SEQ  = 3;
  //int RE   = 4;
  // naming of elements
  int B0     = 0;
  int Ctilde = 1;
  int A10    = 2;
  int A1     = 3;
  int C      = 4;
  int A21    = 5;
  int A2     = 6;

  int L0 = pbm->M * ( 1+pbm->E1 ) * ( 1+pbm->E2 );
  int L = pbm->M * ( 1+pbm->nu*pbm->E1 ) * ( 1+pbm->E2 );

  fprintf( stdout, "Block dimensions: L0 = %d, L = %d\n", L0, L );

  int stateBuffer[5];
  int nextBuffer[5];
  sp->firstState(stateBuffer);

  int idx = 0;
  do {
    int idy = 0;
    sp->firstState(nextBuffer);

    do {

      double valGen = gen->getEntry( idx, idy );
      double valBlock = 0.0;
      int blockId = -1;

      // super-diagonal block
      if ( nextBuffer[DATA] == stateBuffer[DATA] + 1 ) {
        if ( stateBuffer[DATA] == 0 ) {
          valBlock = block[Ctilde]->getEntry( IDX(idx), IDX(idy) );
          blockId = Ctilde;
        }
        else {
          valBlock = block[C]->getEntry( IDX(idx), IDX(idy) );
          blockId = C;
        }
      }
      // infra-diagonal block
      else if ( nextBuffer[DATA] == stateBuffer[DATA] - 1 ) {
        if ( nextBuffer[DATA] == 0 ) {
          valBlock = block[A10]->getEntry( IDX(idx), IDX(idy) );
          blockId = A10;
        }
        else {
          valBlock = block[A21]->getEntry( IDX(idx), IDX(idy) );
          blockId = A21;
        }
      }
      // diagonal block
      else if ( nextBuffer[DATA] == stateBuffer[DATA] ) {
        if ( stateBuffer[DATA] == 0 ) {
          valBlock = block[B0]->getEntry( IDX(idx), IDX(idy) );
          blockId = B0;
        }
        else if ( stateBuffer[DATA] == pbm->N ) {
          valBlock = block[A2]->getEntry( IDX(idx), IDX(idy) );
          blockId = A2;
        }
        else {
          valBlock = block[A1]->getEntry( IDX(idx), IDX(idy) );
          blockId = A1;
        }
      }

      if ( fabs( valGen - valBlock ) > 1e-10 ) {
        fprintf( stdout, "Value mismatch in transition " );
        sp->printState( stdout, stateBuffer );
        fprintf( stdout, " -> " );
        sp->printState( stdout, nextBuffer );
        fprintf( stdout, " Q[%d,%d] != B[%d][%d,%d]:",
                 idx, idy, blockId, IDX(idx), IDX(idy) );
        fprintf( stdout, " %12.8f <> %12.8f (diff=%e)\n",
                 valGen, valBlock, valGen-valBlock );
      }

      sp->nextState( nextBuffer );
      idy++;
    } while(!sp->isZero(nextBuffer));

    sp->nextState( stateBuffer );
    idx++;
  } while (!sp->isZero(stateBuffer));

}

void handleMarginals( problemData* pbm, layeredStateSpace* sp, discreteDistribution* dis, string prefix )
{
  FILE *out;
  string fileName;

  fileName = prefix + ".dat";
  out = fopen( fileName.c_str(), "w" );

  if ( NULL == out ) {
    fprintf( stderr, "Error in handleMarginals(): could not open file '%s' for writing\n",
             fileName.c_str() );
  }
  else {
    // computing marginals and depletion proba
    double **marg = (double**)malloc( 5*sizeof(double*) );
    marg[0] = (double*)calloc( pbm->N+1, sizeof(double) );
    marg[1] = (double*)calloc( pbm->nu+1, sizeof(double) );
    marg[2] = (double*)calloc( pbm->E1+1, sizeof(double) );
    marg[3] = (double*)calloc( pbm->E2+1, sizeof(double) );
    marg[4] = (double*)calloc( pbm->M, sizeof(double) );
    double DP = 0.0;

    int *stateBuffer = (int*)calloc( 5, sizeof(int) );
    sp->firstState(stateBuffer);
    for ( int i = 0; i < sp->cardinal(); i++ ) {
      for( int j = 0; j < 5; j++ ) {
        marg[j][stateBuffer[j]] += dis->getProba(i);
      }
      if ( stateBuffer[2] == 0 ) {
        DP += dis->getProba(i);
      }
      sp->nextState(stateBuffer);
    }

    // computing expected marginals
    double EPQ = 0.0;
    double EDEQ = 0.0;
    double ESEQ = 0.0;
    for ( int i = 0; i <= pbm->N; i++ ) EPQ += i*marg[0][i];
    for ( int i = 0; i <= pbm->E1; i++ ) EDEQ += i*marg[2][i];
    for ( int i = 0; i <= pbm->E2; i++ ) ESEQ += i*marg[3][i];

    // printing out results
    fprintf( out, "# Marginals of the packet queue\n" );
    fprintf( out, "# Average packet length:           %12.6f\n", EPQ );
    fprintf( out, "# col 1: number of packets\n" );
    fprintf( out, "# col 2: probability\n" );
    fprintf( out, "#\n" );
    for ( int i = 0; i <= pbm->N; i++ )
      fprintf( out, "%6d %12e\n", i, marg[0][i] );

    fprintf( out, "\n\n" );

    fprintf( out, "# Marginals of the dominant energy queue\n" );
    fprintf( out, "# Average queue length:           %12.6f\n", EDEQ );
    fprintf( out, "# Depletion probability:          %12.8f\n", DP );
    fprintf( out, "# col 1: number of energy units\n" );
    fprintf( out, "# col 2: probability\n" );
    fprintf( out, "#\n" );
    for ( int i = 0; i <= pbm->E1; i++ )
      fprintf( out, "%6d %12e\n", i, marg[2][i] );

    fprintf( out, "\n\n" );

    fprintf( out, "# Marginals of the secondary energy queue\n" );
    fprintf( out, "# Average queue length:           %12.6f\n", ESEQ );
    fprintf( out, "# col 1: number of energy units\n" );
    fprintf( out, "# col 2: probability\n" );
    fprintf( out, "#\n" );
    for ( int i = 0; i <= pbm->E2; i++ )
      fprintf( out, "%6d %12e\n", i, marg[3][i] );

    fprintf( out, "\n\n" );

    fprintf( out, "# Marginals of the environment\n" );
    fprintf( out, "# col 1: environment state\n" );
    fprintf( out, "# col 2: probability\n" );
    fprintf( out, "#\n" );
    for ( int i = 0; i < pbm->M; i++ )
      fprintf( out, "%6d %12e\n", i, marg[4][i] );

    // cleanup
    free( marg[0] ); free( marg[1] ); free( marg[2] ); free( marg[3] ); free( marg[4] );
    free( marg );
    free( stateBuffer );
    fclose( out );
  }
}

double getDP( problemData* pbm, layeredStateSpace* sp, discreteDistribution* dis )
{
  // computing depletion proba
  double DP = 0.0;

  int *stateBuffer = (int*)calloc( 5, sizeof(int) );
  sp->firstState(stateBuffer);
  for ( int i = 0; i < sp->cardinal(); i++ ) {
    if ( stateBuffer[2] == 0 ) {
      DP += dis->getProba(i);
    }
    sp->nextState(stateBuffer);
  }

  free( stateBuffer );

  return DP;
}

void run_EPEW_experiment( problemData* pbm, layeredStateSpace* sp )
{
  int nbSamples = 10;

  // Running DP experiment with varying data rate
  pbm->dataRate[0][1] = 0.8;
  FILE* out = fopen( "experiment1.dat", "w" );
  if ( NULL == out ) {
    fprintf( stderr, "Error in run_EPEW_experiment(): could not open file '%s' for writing. Using stdout.",
             "experiment1.dat" );
    out = stdout;
  }

  fprintf( out, "# EPEW DP experiment with varying arrival rate\n");
  fprintf( out, "#\n" );
  fprintf( out, "# - column 1: arrival rate lambda_{p,0}^{(1)}\n" );
  fprintf( out, "# - column 2: depletion probability DP\n" );
  fprintf( out, "#\n" );

  for ( int i = 1; i <= nbSamples; i++ ) {
    double lambda = 4.0*i/nbSamples;
    pbm->dataRate[0][0] = lambda;

    sparseMatrix* gen = makeGenerator( sp, pbm );
    markovChain* myMC = new markovChain( gen );
    markovChain* uMC = myMC->uniformize();
    Distribution *stDis0 = uMC->stationaryDistribution_iterative( "Power", 10000, 0.0, "Zero", NULL, /* progress? */ false );
    discreteDistribution *stDis = dynamic_cast<discreteDistribution*>(stDis0);

    double dp = getDP( pbm, sp, stDis );
    fprintf( out, "%8.4f %12.8f\n", lambda, dp );
    fflush( out );

    string name = "marg" + std::to_string(i);
    handleMarginals( pbm, sp, stDis, name );

    // cleanup
    delete stDis0;
    delete uMC;
    delete myMC;
  }

}

int main() {

  problemData *pbm = (problemData*)malloc( sizeof(problemData) );

  pbm->N = 15;
  pbm->E1 = 10;
  pbm->E2 = 2;
  pbm->M = 2;
  pbm->nu = 2;
  pbm->K = 2;

  layeredStateSpace* sp = new layeredStateSpace( pbm->N, pbm->E1, pbm->E2, pbm->M, pbm->nu );
  // std::cout << "Cardinal = " << sp->cardinal() << std::endl << std::endl;
  // sp->enumerate();

  initData_EPEW_Paper( pbm );

  // int experiment = 0; // test experiment
  // int experiment = 1; // M/M/1 experiment
  // int experiment = 2; // EPEW paper experiment
  // int experiment = 3; // submatrix test experiment
  // int experiment = 4; // big battery experiment
  int experiment = 5; // product form experiment

  switch ( experiment ) {
  case 0: {
    // running test & debugging experiment

    sparseMatrix* gen = makeGenerator( sp, pbm );

    /* Debugging the matrix...
  gen->diagnose( stdout );
  string name = "green.map";
  FILE* out = fopen( name.c_str(), "w" );
  fprintf( out, "Pg := Matrix( %ld, {\n", sp->cardinal() );
  gen->write( out, "Maple" );
  fprintf( out, "}, fill=0 );\n");
  fclose(out);
  */

    markovChain* myMC = new markovChain( gen );
    // myMC->write( "Ers", "green" );
  
    // Simulation experiment
    simulationResult* simRes =
        myMC->simulateChain( 1.0e6, /* stats? */ true, /* traj? */ false, /* increments? */ false, /* print? */ false);
    discreteDistribution *simStDis = simRes->getDistribution();

    // Invoking the stationaryDistribution() method forces some parameters to unadapted default
    // Distribution *stDis0 = myMC->stationaryDistribution( /* progress? */ true );
    markovChain* uMC = myMC->uniformize();
    Distribution *stDis0 = uMC->stationaryDistribution_iterative( "Power", 10000, 0.0, "Zero", NULL, /* progress? */ false );
    discreteDistribution *stDis = dynamic_cast<discreteDistribution*>(stDis0);
    // stDis->fprint();

    double dist = stDis->distanceL1( simStDis );
    fprintf( stdout, "Distance between distribs: %10f\n", dist );

    handleMarginals( pbm, sp, simStDis, "simulation" );
    handleMarginals( pbm, sp, stDis, "numerical" );

    delete( stDis0 );
    delete( simRes );
    delete( myMC );
    delete( uMC );
    delete( sp );
  }
    break;

  case 1: {

    initData_MM1( pbm );
    sparseMatrix* gen = makeGenerator( sp, pbm );
    gen->diagnose( stdout );
    markovChain* myMC = new markovChain( gen );
    markovChain* uMC = myMC->uniformize();
    uMC->write( "Ers", "unif_mm1" );
    Distribution *stDis0 = uMC->stationaryDistribution_iterative( "Power", 10000, 0.0, "Zero", NULL, /* progress? */ false );
    discreteDistribution *stDis = dynamic_cast<discreteDistribution*>(stDis0);

    handleMarginals( pbm, sp, stDis, "MM1" );

    delete( stDis0 );
    delete( myMC );
    delete( uMC );
    delete( sp );
  }
    break;

  case 2:
    // running production experiment

    run_EPEW_experiment( pbm, sp );

    break;

  case 3: {

    initData_MM1( pbm );
    sparseMatrix** blocks = makeGeneratorElements( sp, pbm );
    for ( int i = 0; i < 7; i++ ) {
      blocks[i]->diagnose( stdout );
    }
    blocks[1]->write( stdout, "Ers" );

    sparseMatrix* gen = makeGenerator( sp, pbm );

    compareRepresentations( pbm, sp, gen, blocks );
  }
    break;

  case 4: {

    pbm->E1 = 100;
    pbm->E2 = 100;

    delete( sp );
    sp = new layeredStateSpace( pbm->N, pbm->E1, pbm->E2, pbm->M, pbm->nu );

    initData_bigBatteries( pbm );
    sparseMatrix* gen = makeGenerator( sp, pbm );
    gen->diagnose( stdout );
    markovChain* myMC = new markovChain( gen );
    markovChain* uMC = myMC->uniformize();
    // uMC->write( "Ers", "unif_bigBat" );
    Distribution *stDis0 = uMC->stationaryDistribution_iterative( "Power", 10000, 0.0, "Zero", NULL, /* progress? */ true );
    discreteDistribution *stDis = dynamic_cast<discreteDistribution*>(stDis0);

    handleMarginals( pbm, sp, stDis, "bigBattery" );

    delete( stDis0 );
    delete( myMC );
    delete( uMC );
    delete( sp );
  }

  case 5:

    pbm->N = 15;
    pbm->E1 = 10;
    pbm->E2 = 10;
    pbm->nu = 1;
    pbm->M = 1;
    pbm->K = 0;
    pbm->lambda_p = 0.4;
    pbm->lambda_e1 = 0.5;
    pbm->lambda_e2 = 0.2;
    pbm->u1 = 0.1;
    pbm->u2 = 0.1;
    pbm->lambda_minus = 0.2;

    delete( sp );
    sp = new layeredStateSpace( pbm->N, pbm->E1, pbm->E2, pbm->M, pbm->nu );

    initData_productForm( pbm );

    sparseMatrix* gen = makeGenerator( sp, pbm );
    gen->diagnose( stdout );
    markovChain* myMC = new markovChain( gen );
    markovChain* uMC = myMC->uniformize();
    // uMC->write( "Ers", "unif_bigBat" );
    Distribution *stDis0 = uMC->stationaryDistribution_iterative( "Power", 20000, 0.0, "Zero", NULL, /* progress? */ true );
    discreteDistribution *stDis = dynamic_cast<discreteDistribution*>(stDis0);

    handleMarginals( pbm, sp, stDis, "productForm-numerical" );

    discreteDistribution *prDis = solveProductForm( pbm, sp );

    handleMarginals( pbm, sp, prDis, "productForm-exact" );

    delete( stDis0 );
    delete( myMC );
    delete( uMC );
    delete( sp );
  }

  deleteData( pbm );
  free( pbm );

  return 0;
}
