 /**
 * \file          psi.h
 * \author        Florian LEVEQUE
 * \author        Minh Quan HO
 * \author        Jean-Marc.Vincent@imag.fr
 * \version       1.0
 * \date          12/07/2012
 * \brief         Define global data structures : Model and function pointers
 */

/* 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef PSI_H

// Version number
#define myVersion 1.0

#define PSI_H	

#define ALGO_MONOTONE 1
#define ALGO_NONMONOTONE 2

	
//EVENT structure
/**
 * \brief      Event
 * \details   Data structure to represent an event.
 */
typedef struct st_evt{
  	char* name_evt;				/*!< Event identifier               */
	int id_evt;				/*!< Event identifier               */
	void* ptr_transition;			/*!< Transition function of event               */
	char* name_transition;			/*!< Transition function of event               */
	double rate;				/*!< Event occurence rate               */
	
	int nb_queues_origin;			/*!< Number of origins of packets               */
	int* from;				/*!< Set of origins               */
	char** name_from;			/*!< Set of string representation of the origin queues*/
	int nb_fct_index_origin;		/*!< Number of origin index function*/
	void** indexFctOrigin;			/*!< Set of origin index functions               */
	char** names_indexFctOrigin;		/*!< Set of names of origin index functions               */
	
	int nb_queues_dest;			/*!< Number of destinations of packets               */
	int* to;				/*!< Set of destinations               */
	char** name_to;				/*!< Set of string representation of the destination queues*/
	int nb_fct_index_dest;			/*!< Number of destination index function*/
	void** indexFctDest;			/*!< Set of destination index functions               */
	char** names_indexFctDest;		/*!< Set of names of destination index functions               */
	int nb_params;				/*!< Number of extra parameters               */
	char** param;				/*!< Extra parameters               */
	int nb_rand;				/*!< Number of priority*/
	int* rand_par;			/*!< Priority list (Used for Load sharing)               */
} T_event;
//Event List structure
/**
 * \brief      Events
 * \details   Data structure to represent set of events.
 */
typedef struct st_events{
  T_event* evts;		/*!< Set of events*/
  int nb_evts;			/*!< Number of events*/
}T_events;

//Queues structure
/**
 * \brief      Queues
 * \details   Data structure to represent queues.
 */
typedef struct st_queues{
    int* mins;			/*!< Capacity minimum of queues*/
    int* maxs;			/*!< Capacity max of queues*/
    char** name_queues;		/*!< Name of all queues */
    int nb_queues;		/*!< Number of queues*/
} T_queues;

//Model structures
/**
 * \brief      Model
 * \details   Data structure to represent model.
 */
typedef struct st_model{
    T_queues* queues;		/*!< Set of queues*/
    T_events* events;		/*!< Set of events*/
}T_model;

//  Output structure
/**
 * \brief      Output
 * \details   Data structure to represent output.
 */
typedef struct st_output{
  char* outputfilename;		/*!< Name of output file*/
  short int showModel;		/*!< Flag rewrite the model into output file*/
  short int showParam;		/*!< Flag rewrite the parameters into output file*/
  short int showsimulTime;	/*!< Flag write the simulation time into output file*/
  void* outputFct;		/*!< Pointer of output function*/
}T_output;

/**
 * \brief      State
 * \details   Generic structure to represent state in different algorithms.
 */
typedef struct st_state{
    int **queues;		/*!< Set of vectors of queue states */
    int nb_vectors;		/*!< Number of vectors*/
    int splitting;		/*!< Flag to check if splitting has been actived*/
} T_state;


/**
 * \brief      Node [Load Sharing]
 * \details   Structure to describe a worksheet (tree) hierarchical topology type for the events of Work Stealing
 */
struct st_hier_node{
	int num; /*!<  number of the node */
	int level; /*!< current level on the tree */
	struct st_hier_node * father; /*!< pointer to the father node */
	int nbsons; /*!< number of children nodes */
	struct st_linked_sons * sons; /*!<  pointer to the list of children */
};
/**
 * \brief      Sons [Load Sharing]
 * \details   Structure to allow to chain the childrens node in hierarchy
 */
struct st_linked_sons{
	struct st_linked_sons * next; /*!< Next children */
	struct st_hier_node * val; /*!< Value of the children*/
};
/**
 * \brief      Neighbors [Load Sharing]
 * \details   Structure to allow to save neighbors in hierarchy
 */
struct st_neighbors {
	int nb_neighbors; /*!< Number of neighbors */
	int * neighbors; /*!< List of neighbors*/
};
/**
 * \brief      Elements 
 * \details   Structure to represent an element in the hash table
 */
typedef struct elem{
      int length; /*!< Current allocate memory length */
      int occupation; /*!< Number of current index in the list */
      int* index; /*!< List of index of vectors */
}elements;

/**
 * \brief      Pointer of transition function
 * \details   Pointer of transition function
 */
typedef T_state* (*T_ptr_transfct)(T_state* state, T_event* evt);


/**
 * \brief      Pointer of reward function
 * \details   Pointer of reward function
 */
typedef short int (*T_ptr_rewardfct)(T_state* states,int log2temps_arret,int *reward,int* state_saved);	


/**
 * \brief      Pointer of stop function
 * \details   Pointer of stop function
 */
typedef short int (*T_ptr_stopfct)();

/**
 * \brief      Pointer of index function
 * \details   Pointer of index function
 */
typedef double (*T_ptr_indexfct)();

/**
 * \brief      Pointer of output function
 * \details   Pointer of output function
 */
typedef void (*T_ptr_outputfct) ();

/**
 * \brief      Pointer of split function
 * \details   Pointer of output function
 */
typedef int (*T_ptr_splitfct) ();
/**
 * \brief      Pointer of initialize split function
 * \details   Pointer of output function
 */
typedef T_state* (*T_ptr_splitInitfct) (T_state* state,int size);
/**
 * \brief      Pointer of envelop function
 * \details   Pointer of output function
 */
typedef T_state* (*T_ptr_envelop)();

/***************************************************************/
/*               For Simple Forward Parallel                   */
/***************************************************************/

/**
 * \brief     Structure : Intervals in which the global trajectory is decomposed
 */
typedef struct interval {
  int id_interval;
  int*** saved_trajectories;
  int last_saved_trajectory;
  int begin, length;
  int* events_list;
}T_interval;

/**
 * \brief      Structure : Data of a task
 */
typedef struct task_data{
  int id_interval;
  int* begin_state;
  int is_master;
  int priority;
}T_task_data;

/**
 * \brief      Structure : an element of a linked list of tasks with linkage in double direction
 */
typedef struct a_task {
  T_task_data data;
  struct a_task* prec;
  struct a_task* next;
}T_task;

/**
 * \brief      Structure : Taskpool
 */
typedef struct taskpool{
  T_task* head;
  T_task* queue;
  int nb_tasks;
}T_task_pool;

/**
 * \brief      Pointer of task popping function
 * \details   Pointer of task popping function
 */
typedef T_task* (*T_ptr_taskpopfct)(T_task_pool* taskpool);	

/******************************************************************************/
/*                         For Catch me if you can                            */
/******************************************************************************/
/**
 * \brief      Structure : State
 * \details    Linked list of state claculed by each core at a moment 
 */
typedef struct State_Bis {
    int* state; 
    int id;
    struct State_Bis* next;
}T_State_Bis;


/**
 * \brief      Structure : Trajectorie
 * \details    Linked list of states who do the core's trajectorie  
 */
typedef struct Trajectorie{
    int* trajectorie;
    int time;
    struct Trajectorie* next;
}T_Trajectorie;


/**
 * \brief      Structure : Traj
 */
typedef struct Traj{
    T_Trajectorie* trajectorie;
    int id_young;
    int id_old;
    struct Traj* next;
    struct Traj* prec; 
}T_Traj;


/**
 * \brief      Structure : Trajpool
 * \details    double linked list to keep in global memory trajectories ever calcul by core
 */
typedef struct Trajpool{
    T_Traj* head;
    T_Traj* queues;
}T_Trajpool;


/**
 * \brief      Structure : Interval_Bis 
 * \details 	Intervals in which the local trajectory is decomposed
 */
typedef struct Interval_Bis{
    int id_interval;
    T_Trajectorie* saved_trajectorie;
    int begin;
    int index_save_state;
}T_Interval_Bis;
	
#endif //PSI_H
